from typing import Dict

import torch
from torch import nn

from video_transformers.utils.extra import class_to_config


class LinearHead(nn.Module):
    """
     (BxF)
       ↓
    Dropout
       ↓
    Linear
    """

    def __init__(self, hidden_size: int, num_classes: int, dropout_p: float = 0.0):
        super(LinearHead, self).__init__()

        self.hidden_size = hidden_size
        self.num_classes = num_classes
        self.dropout_p = dropout_p

        self.dropout = nn.Dropout(dropout_p) if dropout_p != 0 else None
        self.linear = nn.Linear(in_features=hidden_size, out_features=num_classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: BxF
        if self.dropout is not None:
            x = self.dropout(x)
        x = self.linear(x)
        return x

    @property
    def config(self) -> Dict:
        return class_to_config(self)



from typing import List

from torch import nn


def unfreeze_last_n_stages(stages: List[nn.Module], n: int):
    if n == -1:  # dont freeze if -1
        return

    num_stages = len(stages)
    num_stages_to_freeze = num_stages - n
    for i, stage in enumerate(stages):
        if i >= num_stages_to_freeze:
            break
        for param in stage.parameters():
            param.requires_grad = False


def get_num_trainable_params(model: nn.Module):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def get_num_total_params(model: nn.Module):
    return sum(param.numel() for param in model.parameters())



from pathlib import Path
from typing import Any, Dict, List


def generate_labels_table(labels: List[str]) -> str:
    str_1 = """
| Labels |
| :-- |
"""
    str_2 = "\n".join(["| " + label + " |" for label in labels])
    return str_1 + str_2


def generate_dict_to_table(_dict: Dict[str, Any]) -> str:
    if not _dict:
        return "N/A"
    else:
        return "\n".join([f"| {key} | {value} |" for key, value in _dict.items()])


def generate_hf_model_card(
    task: str = "single_label_classification",
    total_model_params: int = None,
    total_trainable_model_params: int = None,
    labels: List[str] = None,
    preprocessor_config: Dict[str, Any] = None,
    backbone_config: Dict[str, Any] = None,
    neck_config: Dict[str, Any] = None,
    head_config: Dict[str, Any] = None,
) -> str:
    return (
        f"""
---
library_name: video-transformers
tags:
- Video Transformers
- video-transformers
- video-classification
- image-classification
---

## Usage

```python
from video_transformers import VideoModel

model = VideoModel.from_pretrained(model_name_or_path)

model.predict(video_path="video.mp4")
>> [{{'filename': "video.mp4", 'predictions': {{'class1': 0.98, 'class2': 0.02}}}}]
```

- Refer to [video-transformers](https://github.com/video-transformers/) for more details.

## Model description

This model is intended to be used for the task of classifying videos. 
A video is an ordered sequence of frames. An individual frame of a video has spatial information whereas a sequence of video frames have temporal information.
This model can predict the following {len(labels)} labels:

{generate_labels_table(labels)}

| Model Details | Value |
| Task | {task} |
| Total Model Params | {str(total_model_params)} |
| Total Trainable Model Params | {str(total_trainable_model_params)} |

| Preprocessor Config | Value |
| :-- | :-- |
"""
        + generate_dict_to_table(preprocessor_config)
        + """

| Backbone Config | Value |
| :-- | :-- |
"""
        + generate_dict_to_table(backbone_config)
        + """
    
| Neck Config | Value |
| :-- | :-- |
"""
        + generate_dict_to_table(neck_config)
        + """

| Head Config | Value |
| :-- | :-- |
"""
        + generate_dict_to_table(head_config)
        + """

Model-card auto generated by [video-transformers](https://github/fcakyon/video-transformers).
"""
    )


def export_hf_model_card(
    export_dir: str,
    task: str = "single_label_classification",
    total_model_params: int = None,
    total_trainable_model_params: int = None,
    labels: List[str] = None,
    preprocessor_config: Dict[str, Any] = None,
    backbone_config: Dict[str, Any] = None,
    neck_config: Dict[str, Any] = None,
    head_config: Dict[str, Any] = None,
) -> str:
    export_path = Path(export_dir) / "README.md"
    # save as readme.md
    with open(export_path, "w") as f:
        f.write(
            generate_hf_model_card(
                task=task,
                total_model_params=total_model_params,
                total_trainable_model_params=total_trainable_model_params,
                labels=labels,
                preprocessor_config=preprocessor_config,
                backbone_config=backbone_config,
                neck_config=neck_config,
                head_config=head_config,
            )
        )


def generate_gradio_app(
    model_path_or_url: str,
    examples: List[str],
    author_username: str = None,
) -> str:
    from video_transformers import VideoModel

    model = VideoModel.from_pretrained(model_path_or_url)

    return f"""
import gradio as gr

from video_transformers import VideoModel

model: VideoModel = VideoModel.from_pretrained("{model_path_or_url}")

app = gr.Blocks()

with app:
    gr.Markdown("# **<p align='center'>Video Classification with Transformers</p>**")
    gr.Markdown("This space demonstrates the use of hybrid Transformer-based models for video classification.")
    gr.Markdown(f"The model is trained to classify videos belonging to the following classes: {model.labels}")

    with gr.Tabs():
        with gr.TabItem("Upload & Predict"):
            with gr.Box():

                with gr.Row():
                    input_video = gr.Video(label="Input Video", show_label=True)
                    output_label = gr.Label(label="Model Output", show_label=True)

            gr.Markdown("**Predict**")

            with gr.Box():
                with gr.Row():
                    submit_button = gr.Button("Submit")

            gr.Markdown("**Examples:**")

            # gr.Markdown("CricketShot, PlayingCello, Punch, ShavingBeard, TennisSwing")

            with gr.Column():
                gr.Examples({examples}, [input_video], [output_label], model.predict, cache_examples=True)

            submit_button.click(model.predict, inputs=input_video, outputs=[output_label])

            gr.Markdown("**Note:** The model is trained to classify videos belonging to the following classes: {model.labels}")

            gr.Markdown("**Credits:**")
            gr.Markdown("This space is powered by [video-transformers]('https://github.com/video-transformers/')")
            gr.Markdown("{"This model is finetuned by '" + author_username+"'." if author_username else ""}")

app.launch()
"""


if __name__ == "__main__":
    from video_transformers import VideoModel

    model: VideoModel = VideoModel.from_pretrained("runs/hf_exp15/checkpoint")
    print(
        generate_hf_model_card(
            labels=model.labels,
            backbone_config=model.config["backbone"],
            neck_config=model.config["neck"],
            preprocessor_config=model.preprocessor_config,
            head_config=model.config["head"],
            total_model_params=model.num_total_params,
            total_trainable_model_params=model.num_trainable_params,
        )
    )
    print(generate_gradio_app("runs/hf_exp15", examples=["video.mp4"], author_username="fcakyon"))



from typing import Dict


class AutoHead:
    """
    AutoHead is a class that automatically instantiates a video model head from a config.
    """

    @classmethod
    def from_config(cls, config: Dict):
        head_class_name = config.get("name")
        if head_class_name == "LinearHead":
            from video_transformers.heads import LinearHead

            hidden_size = config.get("hidden_size")
            num_classes = config.get("num_classes")
            dropout_p = config.get("dropout_p")
            return LinearHead(hidden_size, num_classes, dropout_p)
        else:
            raise ValueError(f"Unsupported head class name: {head_class_name}")



from argparse import Namespace
from typing import Any, Dict, Generator, List, MutableMapping, Optional

import numpy as np
from torch import Tensor

# modified from lightning/src/pytorch_lightning/utilities/logger.py


def _flatten_dict(params: Dict[Any, Any], delimiter: str = "/") -> Dict[str, Any]:
    """Flatten hierarchical dict, e.g. ``{'a': {'b': 'c'}} -> {'a/b': 'c'}``.
    Args:
        params: Dictionary containing the hyperparameters
        delimiter: Delimiter to express the hierarchy. Defaults to ``'/'``.
    Returns:
        Flattened dict.
    Examples:
        >>> _flatten_dict({'a': {'b': 'c'}})
        {'a/b': 'c'}
        >>> _flatten_dict({'a': {'b': 123}})
        {'a/b': 123}
        >>> _flatten_dict({5: {'a': 123}})
        {'5/a': 123}
    """

    def _dict_generator(
        input_dict: Any, prefixes: List[Optional[str]] = None
    ) -> Generator[Any, Optional[List[str]], List[Any]]:
        prefixes = prefixes[:] if prefixes else []
        if isinstance(input_dict, MutableMapping):
            for key, value in input_dict.items():
                key = str(key)
                if isinstance(value, (MutableMapping, Namespace)):
                    value = vars(value) if isinstance(value, Namespace) else value
                    yield from _dict_generator(value, prefixes + [key])
                else:
                    yield prefixes + [key, value if value is not None else str(None)]
        else:
            yield prefixes + [input_dict if input_dict is None else str(input_dict)]

    return {delimiter.join(keys): val for *keys, val in _dict_generator(params)}


def _sanitize_params(params: Dict[str, Any]) -> Dict[str, Any]:
    """Returns params with non-primitvies converted to strings for logging.
    >>> params = {"float": 0.3,
    ...           "int": 1,
    ...           "string": "abc",
    ...           "bool": True,
    ...           "list": [1, 2, 3],
    ...           "namespace": Namespace(foo=3),
    ...           "layer": torch.nn.BatchNorm1d}
    >>> import pprint
    >>> pprint.pprint(_sanitize_params(params))  # doctest: +NORMALIZE_WHITESPACE
    {'bool': True,
        'float': 0.3,
        'int': 1,
        'layer': "<class 'torch.nn.modules.batchnorm.BatchNorm1d'>",
        'list': '[1, 2, 3]',
        'namespace': 'Namespace(foo=3)',
        'string': 'abc'}
    """
    for k in params.keys():
        # convert relevant np scalars to python types first (instead of str)
        if isinstance(params[k], (np.bool_, np.integer, np.floating)):
            params[k] = params[k].item()
        elif type(params[k]) not in [bool, int, float, str, Tensor]:
            params[k] = str(params[k])
    return params



from pathlib import Path
from typing import List

from video_transformers.templates import generate_gradio_app


def export_gradio_app(
    model,
    examples: List[str],
    author_username: str = None,
    export_dir: str = "runs/exports/",
    export_filename: str = "app.py",
) -> str:
    Path(export_dir).mkdir(parents=True, exist_ok=True)
    app_path = Path(export_dir) / export_filename
    model_dir = str(Path(export_dir) / "checkpoint")
    # save model
    model.save_pretrained(model_dir, config=model.config)
    # save as gradio app
    with open(app_path, "w") as f:
        f.write(generate_gradio_app(model_dir, examples, author_username))



import glob
import os
import re
import urllib.request
from pathlib import Path


def increment_path(path, exist_ok=True, sep=""):
    # Increment path, i.e. runs/exp --> runs/exp{sep}0, runs/exp{sep}1 etc.
    path = Path(path)  # os-agnostic
    if (path.exists() and exist_ok) or (not path.exists()):
        return str(path)
    else:
        dirs = glob.glob(f"{path}{sep}*")  # similar paths
        matches = [re.search(rf"%s{sep}(\d+)" % path.stem, d) for d in dirs]
        i = [int(m.groups()[0]) for m in matches if m]  # indices
        n = max(i) + 1 if i else 2  # increment number
        return f"{path}{sep}{n}"  # update path


def download_file(url: str, download_path: str):
    """
    Downloads a file from the given url to the given path.
    """
    if not os.path.exists(download_path):
        print(f"Downloading {url} to {download_path}")
        urllib.request.urlretrieve(url, download_path)
    else:
        print(f"{download_path} already exists. Skipping download.")



import math
from typing import Dict

import torch
from torch import nn

from video_transformers.utils.extra import class_to_config


class BaseNeck(nn.Module):
    @property
    def config(self) -> Dict:
        return class_to_config(self)


class LSTMNeck(BaseNeck):
    """
        (BxTxF)
           ↓
         LSTM
           ↓
    (BxF) or (BxTxF)
    """

    def __init__(self, num_features, hidden_size, num_layers, return_last=True):
        """
        Create a LSTMNeck.

        Args:
            num_features: Number of input features.
            hidden_size: Number of hidden units.
            num_layers: Number of layers.
            return_last: If True, return the last hidden state of the LSTM.
        """
        super().__init__()

        self.num_features = num_features
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.return_last = return_last

        self.lstm = nn.LSTM(num_features, hidden_size, num_layers, batch_first=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # F1: num_features, F2: hidden_size
        # x: BxTxF1
        x, _ = self.lstm(x)  # x: BxTxF2
        if self.return_last:
            x = x[:, -1, :]  # x: BxF2
        return x


class GRUNeck(BaseNeck):
    """
        (BxTxF)
           ↓
          GRU
           ↓
    (BxF) or (BxTxF)
    """

    def __init__(self, num_features: int, hidden_size: int, num_layers: int, return_last: bool = True):
        """
        Create a GRUNeck.

        Args:
            num_features: Number of input features.
            hidden_size: Number of hidden units.
            num_layers: Number of layers.
            return_last: If True, return the last hidden state of the GRU.
        """
        super().__init__()

        self.num_features = num_features
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.return_last = return_last

        self.gru = nn.GRU(num_features, hidden_size, num_layers, batch_first=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # F1: num_features, F2: hidden_size
        # x: BxTxF1
        x, _ = self.gru(x)  # x: BxTxF2
        if self.return_last:
            x = x[:, -1, :]  # x: BxF2
        return x


class PostionalEncoder(nn.Module):
    """
         (BxTxF)
            ↓
    PostionalEncoder
            ↓
         (BxTxF)
    """

    def __init__(self, num_features: int, dropout_p: float = 0.0, num_timesteps: int = 30):
        super(PostionalEncoder, self).__init__()

        self.num_features = num_features
        self.dropout_p = dropout_p
        self.num_timesteps = num_timesteps

        self.dropout = nn.Dropout(dropout_p) if dropout_p != 0 else None

        self.scale_constat = torch.sqrt(torch.tensor(self.num_features))

        position_encodings = torch.zeros(self.num_timesteps, self.num_features)  # pe: TxF
        for time_ind in range(self.num_timesteps):
            for feat_ind in range(0, self.num_features, 2):
                sin_input = time_ind / (10000 ** ((2 * feat_ind) / self.num_features))
                cos_input = time_ind / (10000 ** ((2 * (feat_ind + 1)) / self.num_features))
                position_encodings[time_ind, feat_ind] = torch.sin(torch.tensor(sin_input))
                position_encodings[time_ind, feat_ind + 1] = torch.cos(torch.tensor(cos_input))

        self.position_encodings = position_encodings

    def add_positional_encoding(self, x: torch.Tensor) -> torch.Tensor:
        # x: BxTxF
        self.position_encodings = self.position_encodings.to(x.device)
        batch_size = x.size(0)
        x = x + self.position_encodings.repeat(batch_size, 1, 1)  # pe: BxTxF
        return x

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: BxTxF

        # end of section 3.4 of 'Attention is All You Need' paper
        x = x * self.scale_constat  # x: BxTxF

        x = self.add_positional_encoding(x)  # x: BxTxF
        if self.dropout is not None:
            x = self.dropout(x)  # x: BxTxF
        return x


class TransformerNeck(BaseNeck):
    """
        (BxTxF)
           ↓
      Transformer
           ↓
    (BxF) or (BxTxF)
    """

    def __init__(
        self,
        num_features: int,
        num_timesteps: int,
        transformer_enc_num_heads: int = 4,
        transformer_enc_num_layers: int = 2,
        transformer_enc_act: int = "gelu",
        dropout_p: int = 0.0,
        return_mean: bool = True,
    ):
        """
        Create a TransformerNeck.

        Args:
            num_features: number of input features
            num_timesteps: number of timesteps
            transformer_enc_num_heads: number of heads in the transformer encoder
            transformer_enc_num_layers: number of layers in the transformer encoder
            transformer_enc_act: activation function for the transformer encoder
            dropout_p: dropout probability
            return_mean: return the mean of the transformed features
        """

        super(TransformerNeck, self).__init__()

        self.num_features = num_features
        self.num_timesteps = num_timesteps
        self.transformer_enc_num_heads = transformer_enc_num_heads
        self.transformer_enc_num_layers = transformer_enc_num_layers
        self.transformer_enc_act = transformer_enc_act
        self.dropout_p = dropout_p
        self.return_mean = return_mean

        self.positional_encoder = PostionalEncoder(num_features, dropout_p, num_timesteps)
        self.encoder_layer = nn.TransformerEncoderLayer(
            d_model=num_features,
            nhead=transformer_enc_num_heads,
            activation=transformer_enc_act,
        )
        self.transformer_encoder = nn.TransformerEncoder(self.encoder_layer, num_layers=transformer_enc_num_layers)

    def forward(self, x):
        # F: num_features
        # x: BxTxF
        x = self.positional_encoder(x)  # x: BxTxF
        x = self.transformer_encoder(x)  # x: BxTxF
        if self.return_mean:
            x = x.mean(dim=1)  # x: BxF
        return x



from typing import Dict, Union

from video_transformers.backbones.base import Backbone
from video_transformers.modeling import TimeDistributed


class AutoBackbone:
    """
    AutoBackbone is a class that automatically instantiates a video model backbone from a config.
    """

    @classmethod
    def from_config(cls, config: Dict) -> Union[Backbone, TimeDistributed]:
        backbone_framework = config.get("framework")
        backbone_type = config.get("type")
        backbone_model_name = config.get("model_name")

        if backbone_framework["name"] == "transformers":
            from video_transformers.backbones.transformers import TransformersBackbone

            backbone = TransformersBackbone(model_name=backbone_model_name)
        elif backbone_framework["name"] == "timm":
            from video_transformers.backbones.timm import TimmBackbone

            backbone = TimmBackbone(model_name=backbone_model_name)
        else:
            raise ValueError(f"Unknown framework {backbone_framework}")

        if backbone_type == "2d_backbone":
            from video_transformers.modeling import TimeDistributed

            backbone = TimeDistributed(backbone)
        return backbone



import json
import os
from pathlib import Path
from typing import Dict, List, Optional, Union

import torch
from huggingface_hub.constants import PYTORCH_WEIGHTS_NAME
from huggingface_hub.file_download import hf_hub_download
from huggingface_hub.hub_mixin import PyTorchModelHubMixin
from torch import nn

import video_transformers.backbones.base
import video_transformers.predict
from video_transformers.heads import LinearHead
from video_transformers.hfhub_wrapper.hub_mixin import push_to_hub
from video_transformers.utils.torch import get_num_total_params, get_num_trainable_params


class Identity(nn.Module):
    def __init__(self):
        super(Identity, self).__init__()

    def forward(self, x):
        return x


class TimeDistributed(nn.Module):
    """
    Time x Backbone2D (BxCxTxHxW)
           ↓
         Pool2D
           ↓
        (BxTxF)
    """

    def __init__(self, backbone: video_transformers.backbones.base.Backbone, low_memory=False):
        super(TimeDistributed, self).__init__()
        self.backbone = backbone
        self.low_memory = low_memory

    @property
    def num_features(self):
        return self.backbone.num_features

    @property
    def mean(self):
        return self.backbone.mean

    @property
    def std(self):
        return self.backbone.std

    @property
    def type(self):
        return self.backbone.type

    @property
    def model_name(self):
        return self.backbone.model_name

    @property
    def framework(self):
        return self.backbone.framework

    @property
    def num_trainable_params(self):
        return self.backbone.num_trainable_params

    @property
    def num_total_params(self):
        return self.backbone.num_total_params

    @property
    def config(self):
        return self.backbone.config

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: BxCxTxHxW
        batch_size, num_channels, num_timesteps, height, width = x.size()

        if self.low_memory:
            output_matrix = []
            for timestep in range(num_timesteps):
                x_t = self.backbone(x[:, :, timestep, :, :])  # x_t: BxC
                output_matrix.append(x_t)
            x = torch.stack(output_matrix, dim=1)  # x: BxTxF
            # clear up memory space
            x_t = None
            output_matrix = None
        else:
            x = x.permute((0, 2, 1, 3, 4))  # x: BxTxCxHxW
            x = x.contiguous().view(batch_size * num_timesteps, num_channels, height, width)  # x: (BxT)xCxHxW
            x = self.backbone(x)  # x: (BxT)xF
            x = x.contiguous().view(batch_size, num_timesteps, x.size(1))  # x: BxTxF
        return x


class VideoModel(nn.Module, PyTorchModelHubMixin):
    """
    (BxCxTxHxW)
         ↓
      Backbone
         ↓
       Neck
         ↓
       (BxF)
         ↓
       Head

    """

    @classmethod
    def from_config(cls, config: Dict) -> "VideoModel":
        """
        Loads a model from a config file.
        """
        import video_transformers

        if isinstance(config, str):
            with open(config) as f:
                config = json.load(f)

        backbone = video_transformers.AutoBackbone.from_config(config["backbone"])
        head = video_transformers.AutoHead.from_config(config["head"])
        neck = None
        if config["neck"] is not None:
            neck = video_transformers.AutoNeck.from_config(config["neck"])

        return cls(
            backbone=backbone,
            head=head,
            neck=neck,
            preprocessor_config=config["preprocessor"],
            labels=config["labels"],
            task=config["task"],
        )

    @classmethod
    def _from_pretrained(
        cls,
        model_id,
        revision,
        cache_dir,
        force_download,
        proxies,
        resume_download,
        local_files_only,
        use_auth_token,
        map_location="cpu",
        strict=False,
        **model_kwargs,
    ):

        map_location = torch.device(map_location)

        if os.path.isdir(model_id):
            print("Loading weights from local directory")
            model_file = os.path.join(model_id, PYTORCH_WEIGHTS_NAME)
        else:
            model_file = hf_hub_download(
                repo_id=model_id,
                filename=PYTORCH_WEIGHTS_NAME,
                revision=revision,
                cache_dir=cache_dir,
                force_download=force_download,
                proxies=proxies,
                resume_download=resume_download,
                use_auth_token=use_auth_token,
                local_files_only=local_files_only,
            )
        model = cls.from_config(**model_kwargs)

        state_dict = torch.load(model_file, map_location=map_location)
        model.load_state_dict(state_dict, strict=strict)
        model.eval()

        return model

    def push_to_hub(
        self,
        *,
        repo_path_or_name: Optional[str] = None,
        repo_url: Optional[str] = None,
        commit_message: Optional[str] = "Add model",
        organization: Optional[str] = None,
        private: bool = False,
        api_endpoint: Optional[str] = None,
        use_auth_token: Optional[Union[bool, str]] = None,
        git_user: Optional[str] = None,
        git_email: Optional[str] = None,
        config: Optional[dict] = None,
        skip_lfs_files: bool = False,
        repo_id: Optional[str] = None,
        token: Optional[str] = None,
        branch: Optional[str] = None,
        create_pr: Optional[bool] = None,
        allow_patterns: Optional[Union[List[str], str]] = None,
        ignore_patterns: Optional[Union[List[str], str]] = None,
    ) -> str:
        """
        Upload model checkpoint to the Hub.

        Use `allow_patterns` and `ignore_patterns` to precisely filter which files
        should be pushed to the hub. See [`upload_folder`] reference for more details.

        Parameters:
            repo_id (`str`, *optional*):
                Repository name to which push.
            commit_message (`str`, *optional*):
                Message to commit while pushing.
            private (`bool`, *optional*, defaults to `False`):
                Whether the repository created should be private.
            api_endpoint (`str`, *optional*):
                The API endpoint to use when pushing the model to the hub.
            token (`str`, *optional*):
                The token to use as HTTP bearer authorization for remote files.
                If not set, will use the token set when logging in with
                `transformers-cli login` (stored in `~/.huggingface`).
            branch (`str`, *optional*):
                The git branch on which to push the model. This defaults to
                the default branch as specified in your repository, which
                defaults to `"main"`.
            create_pr (`boolean`, *optional*):
                Whether or not to create a Pull Request from `branch` with that commit.
                Defaults to `False`.
            config (`dict`, *optional*):
                Configuration object to be saved alongside the model weights.
            allow_patterns (`List[str]` or `str`, *optional*):
                If provided, only files matching at least one pattern are pushed.
            ignore_patterns (`List[str]` or `str`, *optional*):
                If provided, files matching any of the patterns are not pushed.

        Returns:
            The url of the commit of your model in the given repository.
        """
        return push_to_hub(
            self=self,
            repo_path_or_name=repo_path_or_name,
            repo_url=repo_url,
            commit_message=commit_message,
            organization=organization,
            private=private,
            api_endpoint=api_endpoint,
            use_auth_token=use_auth_token,
            git_user=git_user,
            git_email=git_email,
            config=config,
            skip_lfs_files=skip_lfs_files,
            repo_id=repo_id,
            token=token,
            branch=branch,
            create_pr=create_pr,
            allow_patterns=allow_patterns,
            ignore_patterns=ignore_patterns,
        )

    def __init__(
        self,
        backbone: Union[TimeDistributed, video_transformers.backbones.base.Backbone],
        head: LinearHead,
        neck=None,
        labels: List[str] = None,
        preprocessor_config: Dict = None,
        task: str = None,
    ):
        """
        Args:
            backbone: Backbone model.
            head: Head model.
            neck: Neck model.
            labels: List of labels (required for onnx export and predict).
            preprocessor_config: Preprocessor config (required for onnx export and predict).
            task: Task name (required for predict).
        """
        super().__init__()
        self.backbone = backbone
        self.neck = neck
        self.head = head

        # required for exporting to ONNX and predict
        self.preprocessor_config = preprocessor_config
        self.labels = labels

        self.task = task

    @property
    def num_features(self):
        return self.backbone.num_features

    @property
    def mean(self):
        return self.backbone.mean

    @property
    def std(self):
        return self.backbone.std

    @property
    def backbone_type(self):
        return self.backbone.type

    @property
    def backbone_name(self):
        return self.backbone.model_name

    @property
    def backbone_framework(self):
        return self.backbone.framework

    @property
    def num_trainable_params(self):
        return get_num_trainable_params(self)

    @property
    def num_total_params(self):
        return get_num_total_params(self)

    @property
    def config(self):
        config = {}
        config["task"] = self.task
        config["backbone"] = self.backbone.config
        config["head"] = self.head.config
        if self.neck is not None:
            config["neck"] = self.neck.config
        else:
            config["neck"] = None
        config["labels"] = self.labels
        config["preprocessor"] = self.preprocessor_config
        return config

    def to_onnx(
        self,
        quantize: bool = False,
        opset_version: int = 12,
        export_dir: str = "exports/",
        export_filename: str = "model.onnx",
    ):
        """
        Export model to ONNX format.

        Args:
            quantize: Whether to quantize the model.
            opset_version: Opset version to use.
            export_dir: Directory to export model to.
            export_filename: Filename to export model to.
        """
        import video_transformers.deployment.onnx

        video_transformers.deployment.onnx.export(self, quantize, opset_version, export_dir, export_filename)

    def to_gradio(
        self,
        examples: List[str],
        author_username: str = None,
        export_dir: str = "runs/exports/",
        export_filename: str = "app.py",
    ):
        """
        Export model as Gradio App.

        Args:
            examples: List of examples to use for the app.
            author_username: Author's username.
            export_dir: Directory to export model to.
            export_filename: Filename to export model to.
        """
        import video_transformers.deployment.gradio

        video_transformers.deployment.gradio.export_gradio_app(
            self, examples, author_username, export_dir, export_filename
        )

    @property
    def example_input_array(self):
        if self.preprocessor_config:
            return torch.rand(
                1,
                3,
                self.preprocessor_config["num_timesteps"],
                self.preprocessor_config["input_size"],
                self.preprocessor_config["input_size"],
            )
        else:
            return None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: BxCxTxHxW
        x = self.backbone(x)
        if self.neck is not None:
            x = self.neck(x)
        x = self.head(x)
        return x

    def predict(self, video_or_folder_path: Union[str, Path], mode) -> List[Dict]:
        """
        Predict the labels and probabilities of a video or folder of videos.
        Supports local file path/folder directory, S3 URI and https URL.

        Args:
            model: The model to use for prediction.
            video_or_folder_path: The path to the video or folder of videos.
                Supports local file path/folder directory, S3 URI and https URL.
            mode: The mode to use for prediction. Can be "first_batch", "average_all", "random_batch", "uniform_batch".
        """
        result = video_transformers.predict.predict(
            self,
            video_or_folder_path=video_or_folder_path,
            mode=mode,
            preprocessor_config=self.preprocessor_config,
            labels=self.labels,
            device=next(self.parameters()).device,
        )

        return result



class TaskMixin:
    def training_step(self, batch):
        raise NotImplementedError()

    def on_training_epoch_end(self):
        raise NotImplementedError()

    def validation_step(self, batch):
        raise NotImplementedError()

    def on_validation_epoch_end(self):
        raise NotImplementedError()

    @property
    def train_metrics(self):
        raise NotImplementedError()

    @property
    def val_metrics(self):
        raise NotImplementedError()

    @property
    def last_train_result(self):
        return None

    @property
    def last_val_result(self):
        return None



import inspect
from typing import Dict

from torch import nn

from video_transformers.utils.extra import class_to_config
from video_transformers.utils.torch import get_num_total_params, get_num_trainable_params


class Backbone(nn.Module):
    def __init__(
        self,
    ):
        super(Backbone, self).__init__()

    def forward(self, x):
        raise NotImplementedError()

    def unfreeze_last_n_stages(self, n):
        raise NotImplementedError()

    @property
    def num_trainable_params(self):
        return get_num_trainable_params(self.model)

    @property
    def num_total_params(self):
        return get_num_total_params(self.model)

    @property
    def type(self) -> str:
        NotImplementedError()  # 2d_backbone, 3d_backbone

    @property
    def framework(self) -> Dict:
        NotImplementedError()

    @property
    def model_name(self) -> str:
        NotImplementedError()

    @property
    def config(self) -> Dict:
        return class_to_config(self)



import os
import tempfile
from pathlib import Path
from typing import List, Optional, Union

from huggingface_hub import hf_api
from huggingface_hub.hf_api import HfApi, HfFolder
from huggingface_hub.repository import Repository

from video_transformers.templates import export_hf_model_card


def push_to_hub(
    self,
    # NOTE: deprecated signature that will change in 0.12
    *,
    repo_path_or_name: Optional[str] = None,
    repo_url: Optional[str] = None,
    commit_message: Optional[str] = "Add model",
    organization: Optional[str] = None,
    private: bool = False,
    api_endpoint: Optional[str] = None,
    use_auth_token: Optional[Union[bool, str]] = None,
    git_user: Optional[str] = None,
    git_email: Optional[str] = None,
    config: Optional[dict] = None,
    skip_lfs_files: bool = False,
    # NOTE: New arguments since 0.9
    repo_id: Optional[str] = None,  # optional only until 0.12
    token: Optional[str] = None,
    branch: Optional[str] = None,
    create_pr: Optional[bool] = None,
    allow_patterns: Optional[Union[List[str], str]] = None,
    ignore_patterns: Optional[Union[List[str], str]] = None,
    # TODO (release 0.12): signature must be the following
    # repo_id: str,
    # *,
    # commit_message: Optional[str] = "Add model",
    # private: bool = False,
    # api_endpoint: Optional[str] = None,
    # token: Optional[str] = None,
    # branch: Optional[str] = None,
    # create_pr: Optional[bool] = None,
    # config: Optional[dict] = None,
    # allow_patterns: Optional[Union[List[str], str]] = None,
    # ignore_patterns: Optional[Union[List[str], str]] = None,
) -> str:
    """
    Upload model checkpoint to the Hub.

    Use `allow_patterns` and `ignore_patterns` to precisely filter which files
    should be pushed to the hub. See [`upload_folder`] reference for more details.

    Parameters:
        repo_id (`str`, *optional*):
            Repository name to which push.
        commit_message (`str`, *optional*):
            Message to commit while pushing.
        private (`bool`, *optional*, defaults to `False`):
            Whether the repository created should be private.
        api_endpoint (`str`, *optional*):
            The API endpoint to use when pushing the model to the hub.
        token (`str`, *optional*):
            The token to use as HTTP bearer authorization for remote files.
            If not set, will use the token set when logging in with
            `transformers-cli login` (stored in `~/.huggingface`).
        branch (`str`, *optional*):
            The git branch on which to push the model. This defaults to
            the default branch as specified in your repository, which
            defaults to `"main"`.
        create_pr (`boolean`, *optional*):
            Whether or not to create a Pull Request from `branch` with that commit.
            Defaults to `False`.
        config (`dict`, *optional*):
            Configuration object to be saved alongside the model weights.
        allow_patterns (`List[str]` or `str`, *optional*):
            If provided, only files matching at least one pattern are pushed.
        ignore_patterns (`List[str]` or `str`, *optional*):
            If provided, files matching any of the patterns are not pushed.

    Returns:
        The url of the commit of your model in the given repository.
    """
    # If the repo id is set, it means we use the new version using HTTP endpoint
    # (introduced in v0.9).
    if repo_id is not None:
        token, _ = hf_api._validate_or_retrieve_token(token)
        api = HfApi(endpoint=api_endpoint)

        api.create_repo(
            repo_id=repo_id,
            repo_type="model",
            token=token,
            private=private,
            exist_ok=True,
        )

        # Push the files to the repo in a single commit
        with tempfile.TemporaryDirectory() as tmp:
            saved_path = Path(tmp) / repo_id
            self.save_pretrained(saved_path, config=config)
            export_hf_model_card(
                export_dir=saved_path,
                labels=self.labels,
                backbone_config=self.config["backbone"],
                neck_config=self.config["neck"],
                preprocessor_config=self.config["preprocessor"],
                head_config=self.config["head"],
                total_model_params=self.num_total_params,
                total_trainable_model_params=self.num_trainable_params,
            )
            return api.upload_folder(
                repo_id=repo_id,
                repo_type="model",
                token=token,
                folder_path=saved_path,
                commit_message=commit_message,
                revision=branch,
                create_pr=create_pr,
                allow_patterns=allow_patterns,
                ignore_patterns=ignore_patterns,
            )

    # If the repo id is None, it means we use the deprecated version using Git
    # TODO: remove code between here and `return repo.git_push()` in release 0.12
    if repo_path_or_name is None and repo_url is None:
        raise ValueError("You need to specify a `repo_path_or_name` or a `repo_url`.")

    if use_auth_token is None and repo_url is None:
        token = HfFolder.get_token()
        if token is None:
            raise ValueError(
                "You must login to the Hugging Face hub on this computer by typing"
                " `huggingface-cli login` and entering your credentials to use"
                " `use_auth_token=True`. Alternatively, you can pass your own token"
                " as the `use_auth_token` argument."
            )
    elif isinstance(use_auth_token, str):
        token = use_auth_token
    else:
        token = None

    if repo_path_or_name is None:
        repo_path_or_name = repo_url.split("/")[-1]

    # If no URL is passed and there's no path to a directory containing files, create a repo
    if repo_url is None and not os.path.exists(repo_path_or_name):
        repo_id = Path(repo_path_or_name).name
        if organization:
            repo_id = f"{organization}/{repo_id}"
        repo_url = HfApi(endpoint=api_endpoint).create_repo(
            repo_id=repo_id,
            token=token,
            private=private,
            repo_type=None,
            exist_ok=True,
        )

    repo = Repository(
        repo_path_or_name,
        clone_from=repo_url,
        use_auth_token=use_auth_token,
        git_user=git_user,
        git_email=git_email,
        skip_lfs_files=skip_lfs_files,
    )
    repo.git_pull(rebase=True)

    # Save the files in the cloned repo
    self.save_pretrained(repo_path_or_name, config=config)
    export_hf_model_card(
        export_dir=saved_path,
        labels=self.labels,
        backbone_config=self.config["backbone"],
        neck_config=self.config["neck"],
        preprocessor_config=self.config["preprocessor"],
        head_config=self.config["head"],
        total_model_params=self.num_total_params,
        total_trainable_model_params=self.num_trainable_params,
    )

    # Commit and push!
    repo.git_add(auto_lfs_track=True)
    repo.git_commit(commit_message)
    return repo.git_push()



import torch


def get_multistep_scheduler_with_warmup(
    optimizer: torch.optim.Optimizer, max_epochs: int = 12, warmup_epochs: float = 0.1
):
    """
    Torch multistep learning rate scheduler with warmup.
    Decrease the learning rate at milestones by a factor of 0.1.
    Milestones are chosen as 7/10 and 9/10 of total epochs.
    """
    warmup_milestone_epoch = max(1.0, int(warmup_epochs * max_epochs))
    warmup_scheduler = torch.optim.lr_scheduler.LinearLR(
        optimizer,
        start_factor=0.01,
        end_factor=1.0,
        total_iters=max(1.0, warmup_milestone_epoch),
    )

    ms_scheduler = torch.optim.lr_scheduler.MultiStepLR(
        optimizer,
        milestones=[
            round((max_epochs - 1) * 7 / 10),
            round((max_epochs - 1) * 9 / 10),
        ],
        gamma=0.1,
        last_epoch=-1,
    )

    scheduler = torch.optim.lr_scheduler.SequentialLR(
        optimizer,
        [warmup_scheduler, ms_scheduler],
        milestones=[warmup_milestone_epoch],
    )

    return scheduler


def get_linear_scheduler_with_warmup(
    optimizer: torch.optim.Optimizer, max_epochs: int = 12, warmup_epochs: float = 0.1
):
    """
    Torch linear learning rate scheduler with warmup.
    """
    warmup_milestone_epoch = max(1.0, int(warmup_epochs * max_epochs))
    warmup_scheduler = torch.optim.lr_scheduler.LinearLR(
        optimizer,
        start_factor=0.01,
        end_factor=1.0,
        total_iters=warmup_milestone_epoch,
    )

    linear_scheduler = torch.optim.lr_scheduler.LinearLR(
        optimizer,
        start_factor=1.0,
        end_factor=0.01,
        total_iters=max_epochs - warmup_milestone_epoch,
    )

    scheduler = torch.optim.lr_scheduler.SequentialLR(
        optimizer,
        [warmup_scheduler, linear_scheduler],
        milestones=[warmup_milestone_epoch],
    )

    return scheduler


def get_cosineannealing_scheduler_with_warmup(
    optimizer: torch.optim.Optimizer, max_epochs: int = 12, warmup_epochs: float = 0.1
):

    warmup_milestone_epoch = max(1.0, int(warmup_epochs * max_epochs))
    warmup_scheduler = torch.optim.lr_scheduler.LinearLR(
        optimizer,
        start_factor=0.01,
        end_factor=1.0,
        total_iters=max(1.0, warmup_milestone_epoch),
    )

    cosineannealing_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
        optimizer,
        T_max=max_epochs - warmup_milestone_epoch,
    )

    scheduler = torch.optim.lr_scheduler.SequentialLR(
        optimizer,
        [warmup_scheduler, cosineannealing_scheduler],
        milestones=[warmup_milestone_epoch],
    )

    return scheduler



from typing import Dict

from video_transformers.backbones.base import Backbone
from video_transformers.utils.torch import unfreeze_last_n_stages as unfreeze_last_n_stages_torch

models_2d = ["convnext", "levit", "cvt", "clip", "swin", "vit", "deit", "beit"]
models_3d = ["videomae"]


class TransformersBackbone(Backbone):
    def __init__(self, model_name: str, num_unfrozen_stages=0, **backbone_kwargs):
        super(TransformersBackbone, self).__init__()
        from transformers import AutoFeatureExtractor, AutoModel

        backbone = AutoModel.from_pretrained(model_name, **backbone_kwargs)
        feature_extractor = AutoFeatureExtractor.from_pretrained(model_name, **backbone_kwargs)

        if backbone.base_model_prefix == "clip":
            from transformers import CLIPVisionModel

            backbone = CLIPVisionModel.from_pretrained(model_name)

        mean, std = feature_extractor.image_mean, feature_extractor.image_std

        if hasattr(backbone.config, "hidden_size"):  # vit, swin, deit
            num_features = backbone.config.hidden_size
        elif hasattr(backbone.config, "hidden_sizes"):  # levit, convnext
            num_features = backbone.config.hidden_sizes[-1]
        elif hasattr(backbone.config, "embed_dim"):  # cvt
            num_features = backbone.config.embed_dim[-1]
        elif hasattr(backbone.config, "projection_dim"):  # clip
            num_features = backbone.config.projection_dim
        else:
            raise NotImplementedError(f"Huggingface model not supported: {backbone.base_model_prefix}")

        # set model input num frames
        if hasattr(backbone.config, "num_frames"):  # videomae
            self._num_frames = backbone.config.num_frames
        else:
            self._num_frames = 1

        self.model = backbone
        self.num_features = num_features
        self.mean = mean
        self.std = std
        self._model_name = model_name

        if self.model.base_model_prefix in models_2d:
            self._type = "2d_backbone"
        elif self.model.base_model_prefix in models_3d:
            self._type = "3d_backbone"
        else:
            raise NotImplementedError(f"Huggingface model not supported: {self.model.base_model_prefix}")

        self.unfreeze_last_n_stages(num_unfrozen_stages)

    @property
    def type(self) -> str:
        return self._type

    @property
    def framework(self) -> Dict:
        import transformers

        return {"name": "transformers", "version": transformers.__version__}

    @property
    def model_name(self) -> str:
        return self._model_name

    @property
    def num_frames(self) -> int:
        return self._num_frames

    def forward(self, x):
        if self.model.base_model_prefix in models_3d:
            # ensure num_timesteps matches model num_frames
            if x.shape[2] != self.num_frames:
                raise ValueError(
                    f"Input has {x.shape[2]} frames, but {self.model_name} accepts {self.num_frames} frames. Set num_timesteps to {self.num_frames}."
                )
            # x: (B, C, T, H, W)
            x = x.permute(0, 2, 1, 3, 4)
            # x: (B, T, C, H, W)
            output = self.model(pixel_values=x, return_dict=True)[0]
            # output: batch x latent_dim x num_features
        else:
            output = self.model(pixel_values=x, return_dict=True)[1]
            # output: batch x 1 x num_features
        if output.dim() == 3:
            output = output.mean(1)
        return output  # output: batch x num_features

    def unfreeze_last_n_stages(self, n):
        if self.model.base_model_prefix == "convnext":
            stages = []
            # stages.append(self.model.base_model.embeddings)
            # freeze embeddings
            for param in self.model.base_model.embeddings.parameters():
                param.requires_grad = False
            # unfreeze last n stages
            stages.extend(self.model.base_model.encoder.stages)
            stages.append(self.model.base_model.layernorm)
            unfreeze_last_n_stages_torch(stages, n)
        elif self.model.base_model_prefix == "levit":
            stages = []
            # stages.extend(list(self.model.base_model.patch_embeddings.children()))
            # freeze embeddings
            for param in self.model.base_model.patch_embeddings.parameters():
                param.requires_grad = False
            # unfreeze last n stages
            stages.extend(self.model.base_model.encoder.stages)
            unfreeze_last_n_stages_torch(stages, n)
        elif self.model.base_model_prefix == "cvt":
            stages = []
            stages.extend(list(self.model.base_model.encoder.stages.children()))
            unfreeze_last_n_stages_torch(stages, n)
        elif self.model.base_model_prefix == "clip":
            stages = []
            # stages.extend(list(self.model.base_model.vision_model.embeddings.children()))
            # freeze embeddings
            for param in self.model.base_model.vision_model.embeddings.parameters():
                param.requires_grad = False
            # unfreeze last n stages
            stages.extend(list(self.model.base_model.vision_model.encoder.layers.children()))
            unfreeze_last_n_stages_torch(stages, n)
        elif self.model.base_model_prefix in ["swin", "vit", "deit", "beit"]:
            stages = []
            # stages.extend(list(self.model.base_model.embeddings.children()))
            # freeze embeddings
            for param in self.model.base_model.embeddings.parameters():
                param.requires_grad = False
            # unfreeze last n stages
            stages.extend(list(self.model.base_model.encoder.layers.children()))
            stages.append(self.model.base_model.layernorm)
            unfreeze_last_n_stages_torch(stages, n)
        elif self.model.base_model_prefix == "videomae":
            stages = []
            # freeze embeddings
            for param in self.model.base_model.embeddings.parameters():
                param.requires_grad = False
            # unfreeze last n stages
            stages.extend(list(self.model.base_model.encoder.layer.children()))
            unfreeze_last_n_stages_torch(stages, n)
        else:
            raise NotImplementedError(f"Freezing not supported for Huggingface model: {self.model.base_model_prefix}")



from typing import Any, List

import evaluate
import torch

from video_transformers.tasks.base import TaskMixin


class Combine:
    # place holder for evaluate.combine till https://github.com/huggingface/evaluate/issues/234 fixed
    def __init__(self, metrics: List[str]):
        self.metrics = [evaluate.load(metric) if isinstance(metric, str) else metric for metric in metrics]

    def add_batch(self, predictions: Any, references: Any):
        for metric in self.metrics:
            metric.add_batch(predictions=predictions, references=references)

    def compute(self, **kwargs):
        results = {}
        zero_division = kwargs.get("zero_division", "warn")
        kwargs.pop("zero_division")
        for metric in self.metrics:
            if metric.name == "precision":
                results.update(metric.compute(zero_division=zero_division, **kwargs))
            else:
                results.update(metric.compute(**kwargs))
        return results


class SingleLabelClassificationTaskMixin(TaskMixin):
    def __init__(self, *args, **kwargs):
        super(SingleLabelClassificationTaskMixin, self).__init__(*args, **kwargs)
        self._train_metrics = Combine(["f1", "precision", "recall"])
        self._val_metrics = Combine(["f1", "precision", "recall"])
        self._last_train_result = None
        self._last_val_result = None
        self.task = "single_label_classification"

    @property
    def train_metrics(self):
        return self._train_metrics

    @property
    def val_metrics(self):
        return self._val_metrics

    @property
    def last_train_result(self):
        return self._last_train_result

    @property
    def last_val_result(self):
        return self._last_val_result

    def training_step(self, batch):
        inputs = batch["video"]
        labels = batch["label"]
        outputs = self.model(inputs)
        probabilities = torch.nn.functional.softmax(outputs, dim=1)
        predictions = probabilities.argmax(dim=-1)
        # gather all predictions and targets
        all_predictions = self.accelerator.gather(predictions)
        all_labels = self.accelerator.gather(labels)
        # accumulate for evaluation
        self.train_metrics.add_batch(predictions=all_predictions, references=all_labels)
        # calculate loss
        loss = self.loss_function(outputs, labels)
        return loss

    def on_training_epoch_end(self):
        if self.accelerator.is_main_process:
            result = self.train_metrics.compute(average=None, zero_division=0)
            for metric, scores in result.items():
                # log macro score
                self.accelerator.log({f"train/{metric}": scores.mean()}, step=self.overall_epoch)
                # log label-wise scores
                for label, score in zip(self.labels, scores):
                    self.accelerator.log(
                        {f"train/{label}_{metric}": score},
                        step=self.overall_epoch,
                    )
            self._last_train_result = result

    def validation_step(self, batch):
        inputs = batch["video"]
        labels = batch["label"]
        outputs = self.model(inputs)
        probabilities = torch.nn.functional.softmax(outputs, dim=1)
        predictions = probabilities.argmax(dim=-1)
        # gather all predictions and targets
        all_predictions = self.accelerator.gather(predictions)
        all_labels = self.accelerator.gather(labels)
        # accumulate for evaluation
        self.val_metrics.add_batch(predictions=all_predictions, references=all_labels)
        # calculate loss
        loss = self.loss_function(outputs, labels)
        return loss

    def on_validation_epoch_end(self):
        if self.accelerator.is_main_process:
            result = self.val_metrics.compute(average=None, zero_division=0)
            for metric, scores in result.items():
                # log macro score
                self.accelerator.log({f"val/{metric}": scores.mean()}, step=self.overall_epoch)
                # log label-wise scores
                for label, score in zip(self.labels, scores):
                    self.accelerator.log(
                        {f"val/{label}_{metric}": score},
                        step=self.overall_epoch,
                    )
            self._last_val_result = result



from typing import Dict, Tuple

import pytorchvideo.data
import torch
import torch.utils.data
from accelerate.logging import get_logger
from pytorchvideo.transforms import (
    ApplyTransformToKey,
    Normalize,
    RandomShortSideScale,
    ShortSideScale,
    UniformTemporalSubsample,
)
from torch.utils.data import DataLoader
from torchvision.transforms import CenterCrop, Compose, Lambda, RandomCrop, RandomHorizontalFlip

from video_transformers.pytorchvideo_wrapper.data.labeled_video_paths import LabeledVideoDataset, LabeledVideoPaths
from video_transformers.utils.extra import class_to_config

logger = get_logger(__name__)


class VideoPreprocessor:
    @classmethod
    def from_config(cls, config: Dict, **kwargs) -> "VideoPreprocessor":
        """
        Creates an instance of the class from a config.

        Args:
            config: config dictionary
            kwargs: owerwrite config values
        """
        for ke, val in kwargs.items():
            config.update({ke, val})
        return cls(**config)

    def __init__(
        self,
        num_timesteps: int = 8,
        input_size: int = 224,
        means: Tuple[float] = (0.45, 0.45, 0.45),
        stds: Tuple[float] = (0.225, 0.225, 0.225),
        min_short_side: int = 256,
        max_short_side: int = 320,
        horizontal_flip_p: float = 0.5,
        clip_duration: int = 1,
    ):
        """
        Creates preprocess transforms.

        Args:
            num_timesteps: number of frames in a video clip
            input_size: model input isze
            means: mean of the video clip
            stds: standard deviation of the video clip
            min_short_side_scale: minimum short side of the video clip
            max_short_side_scale: maximum short side of the video clip
            horizontal_flip_p: probability of horizontal flip
            clip_duration: duration of each video clip

        Properties:
            train_transform: transforms for training
            train_video_transform: training transforms for video clips
            val_transform: transforms for validation
            val_video_transform: validation transforms for video clips

        """
        super().__init__()

        self.num_timesteps = num_timesteps
        self.input_size = input_size
        self.means = means
        self.stds = stds
        self.min_short_side = min_short_side
        self.max_short_side = max_short_side
        self.horizontal_flip_p = horizontal_flip_p
        self.clip_duration = clip_duration

        # Transforms applied to train dataset.
        self.train_video_transform = Compose(
            [
                UniformTemporalSubsample(self.num_timesteps),
                Lambda(lambda x: x / 255.0),
                Normalize(self.means, self.stds),
                RandomShortSideScale(
                    min_size=self.min_short_side,
                    max_size=self.max_short_side,
                ),
                RandomCrop(self.input_size),
                RandomHorizontalFlip(p=self.horizontal_flip_p),
            ]
        )

        self.train_transform = ApplyTransformToKey(key="video", transform=self.train_video_transform)

        # Transforms applied on val dataset or for inference.
        self.val_video_transform = Compose(
            [
                UniformTemporalSubsample(self.num_timesteps),
                Lambda(lambda x: x / 255.0),
                Normalize(self.means, self.stds),
                ShortSideScale(self.min_short_side),
                CenterCrop(self.input_size),
            ]
        )
        self.val_transform = ApplyTransformToKey(key="video", transform=self.val_video_transform)


class VideoDataModule:
    def __init__(
        self,
        train_root: str,
        val_root: str,
        test_root: str = None,
        train_dataset_multiplier: int = 1,
        batch_size: int = 4,
        num_workers: int = 4,
        num_timesteps: int = 8,
        preprocess_input_size: int = 224,
        preprocess_clip_duration: int = 1,
        preprocess_means: Tuple[float] = (0.45, 0.45, 0.45),
        preprocess_stds: Tuple[float] = (0.225, 0.225, 0.225),
        preprocess_min_short_side: int = 256,
        preprocess_max_short_side: int = 320,
        preprocess_horizontal_flip_p: float = 0.5,
    ):
        """
        Initialize video data module.

        Folder structure:
            train_root
                label_1
                    video_1
                    video_2
                    ...
                label_2
                    video_1
                    video_2
                    ...
                ...
            val_root
                label_1
                    video_1
                    video_2
                    ...
                label_2
                    video_1
                    video_2
                    ...
                ...

        Args:
            train_root: str
                Path to kinetics formatted train folder.
            val_root: str
                Path to kinetics formatted train folder.
            test_root: str
                Path to kinetics formatted train folder.
            clip_duration: float
                Duration of sampled clip for each video.
            train_dataset_multiplier: int
                Multipler for number of of random training data samples.
            batch_size: int
                Batch size for training and validation.
            num_workers: int
                Number of parallel processes fetching data.
            num_timesteps: int
                Number of frames to subsample from each clip.
            preprocess_crop_size: int
                Size of the random crop.
            preprocess_means: Tuple[float]
                Mean pixel value to be used during normalization.
            preprocess_stds: Tuple[float]
                Standard deviation pixel value to be used during normalization.
            preprocess_min_short_side: int
                Minimum value of the short side of the clip after resizing.
            preprocess_max_short_side: int
                Maximum value of the short side of the clip after resizing.
            preprocess_horizontal_flip_p: float
                Probability of horizontal flip.
        """
        self.preprocessor_config = {
            "num_timesteps": num_timesteps,
            "input_size": preprocess_input_size,
            "means": preprocess_means,
            "stds": preprocess_stds,
            "min_short_side": preprocess_min_short_side,
            "max_short_side": preprocess_max_short_side,
            "horizontal_flip_p": preprocess_horizontal_flip_p,
            "clip_duration": preprocess_clip_duration,
        }
        self.preprocessor = VideoPreprocessor.from_config(self.preprocessor_config)

        self.dataloader_config = {"batch_size": batch_size, "num_workers": num_workers}

        self.train_root = train_root
        self.val_root = val_root
        self.test_root = test_root if test_root is not None else val_root
        self.train_dataset_multiplier = train_dataset_multiplier
        self.labels = None

        self.train_dataloader = self._get_train_dataloader()
        self.val_dataloader = self._get_val_dataloader()
        self.test_dataloader = self._get_test_dataloader()

    @property
    def num_classes(self):
        return len(self.labels)

    @property
    def config(self) -> Dict:
        return class_to_config(self, ignored_attrs=("config", "train_root", "val_root", "test_root"))

    def _get_train_dataloader(self):
        labeled_video_paths = LabeledVideoPaths.from_path(self.train_root)
        labeled_video_paths.path_prefix = ""
        video_sampler = torch.utils.data.RandomSampler
        clip_sampler = pytorchvideo.data.make_clip_sampler("random", self.preprocessor_config["clip_duration"])
        dataset = LabeledVideoDataset(
            labeled_video_paths,
            clip_sampler,
            video_sampler,
            self.preprocessor.train_transform,
            decode_audio=False,
            decoder="pyav",
            dataset_multiplier=self.train_dataset_multiplier,
        )
        self.labels = dataset.labels
        return DataLoader(
            dataset,
            batch_size=self.dataloader_config["batch_size"],
            num_workers=self.dataloader_config["num_workers"],
            drop_last=False,
        )

    def _get_val_dataloader(self):
        labeled_video_paths = LabeledVideoPaths.from_path(self.val_root)
        labeled_video_paths.path_prefix = ""
        video_sampler = torch.utils.data.SequentialSampler
        clip_sampler = pytorchvideo.data.clip_sampling.UniformClipSamplerTruncateFromStart(
            clip_duration=self.preprocessor_config["clip_duration"],
            truncation_duration=self.preprocessor_config["clip_duration"],
        )
        dataset = LabeledVideoDataset(
            labeled_video_paths,
            clip_sampler,
            video_sampler,
            self.preprocessor.val_transform,
            decode_audio=False,
            decoder="pyav",
        )
        return DataLoader(
            dataset,
            batch_size=self.dataloader_config["batch_size"],
            num_workers=self.dataloader_config["num_workers"],
            drop_last=False,
        )

    def _get_test_dataloader(self):
        labeled_video_paths = LabeledVideoPaths.from_path(self.test_root)
        labeled_video_paths.path_prefix = ""
        video_sampler = torch.utils.data.SequentialSampler
        clip_sampler = pytorchvideo.data.clip_sampling.UniformClipSamplerTruncateFromStart(
            clip_duration=self.preprocessor_config["clip_duration"],
            truncation_duration=self.preprocessor_config["clip_duration"],
        )
        dataset = LabeledVideoDataset(
            labeled_video_paths,
            clip_sampler,
            video_sampler,
            self.preprocessor.val_transform,
            decode_audio=False,
            decoder="pyav",
        )
        return DataLoader(
            dataset,
            batch_size=self.dataloader_config["batch_size"],
            num_workers=self.dataloader_config["num_workers"],
            drop_last=False,
        )



from typing import Tuple

from torch import nn

from video_transformers.backbones.base import Backbone
from video_transformers.modeling import Identity
from video_transformers.utils.torch import unfreeze_last_n_stages as unfreeze_last_n_stages_torch


class TimmBackbone(Backbone):
    def __init__(self, model_name: str, pretrained: bool = False, num_unfrozen_stages=0, **backbone_kwargs):
        super(TimmBackbone, self).__init__()
        import timm

        backbone = timm.create_model(model_name, pretrained=pretrained, **backbone_kwargs)
        if backbone.pretrained_cfg["classifier"] == "head.fc":
            backbone.head.fc = Identity()
        elif backbone.pretrained_cfg["classifier"] == "fc":  # adv_inception_v3
            backbone.fc = Identity()
        elif backbone.pretrained_cfg["classifier"] == "head":  # deit
            backbone.head = Identity()
        elif backbone.pretrained_cfg["classifier"] == "classifier":
            backbone.classifier = Identity()
        elif backbone.pretrained_cfg["classifier"] == ("head.l", "head_dist.l"):  # levit
            backbone = nn.Sequential(*list(backbone.children())[:-2])
        else:
            raise NotImplementedError(f"Backbone not supported: {backbone.pretrained_cfg}")

        mean, std = backbone.pretrained_cfg["mean"], backbone.pretrained_cfg["std"]
        num_features = backbone.num_features

        self.model = backbone
        self.num_features = num_features
        self.mean = mean
        self.std = std
        self._model_name = model_name
        self._type = "2d_backbone"

        self.unfreeze_last_n_stages(num_unfrozen_stages)

    @property
    def type(self):
        return self._type

    @property
    def framework(self):
        import timm

        return {"name": "timm", "version": timm.__version__}

    @property
    def model_name(self):
        return self._model_name

    def forward(self, x):
        return self.model(x)

    def unfreeze_last_n_stages(self, n):
        stages = [stage for stage in self.model.stem.children()] + [stage for stage in self.model.stages.children()]
        unfreeze_last_n_stages_torch(stages, n)



import io
import os
import re

import setuptools


def get_long_description():
    base_dir = os.path.abspath(os.path.dirname(__file__))
    with io.open(os.path.join(base_dir, "README.md"), encoding="utf-8") as f:
        return f.read()


def get_requirements():
    with open("requirements.txt") as f:
        return f.read().splitlines()


def get_version():
    current_dir = os.path.abspath(os.path.dirname(__file__))
    version_file = os.path.join(current_dir, "video_transformers", "__init__.py")
    with io.open(version_file, encoding="utf-8") as f:
        return re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)


_DEV_REQUIREMENTS = ["black==21.7b0", "flake8==3.9.2", "isort==5.9.2"]

_TEST_REQUIREMENTS = ["onnx", "onnxruntime"]

extras = {"test": [_DEV_REQUIREMENTS + _TEST_REQUIREMENTS], "dev": _DEV_REQUIREMENTS}


setuptools.setup(
    name="video-transformers",
    version=get_version(),
    author="fcakyon",
    license="MIT",
    description="Easiest way of fine-tuning HuggingFace video classification models.",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/fcakyon/video-transformers",
    packages=setuptools.find_packages(exclude=["examples", "tests"]),
    python_requires=">=3.7",
    install_requires=get_requirements(),
    extras_require=extras,
    include_package_data=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Education",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    keywords="machine-learning, deep-learning, ml, pytorch, vision, loss, video-classification, transformers, accelerate, evaluate, huggingface",
)



from collections import defaultdict
from typing import List

import pytorchvideo.data
import torch
import torch.utils.data
from iopath.common.file_io import g_pathmgr

from video_transformers.data import VideoPreprocessor
from video_transformers.pytorchvideo_wrapper.data.labeled_video_paths import LabeledVideoDataset, LabeledVideoPaths


def predict(
    model,
    video_or_folder_path,
    preprocessor_config: dict,
    labels: List[str],
    mode: str = "first_batch",
    device: str = None,
):
    """
    Predict the labels of a video or folder of videos.
    Supports local file path/folder directory, S3 URI and https URL.

    Args:
        model: The model to use for prediction.
        video_or_folder_path: The path to the video or folder of videos.
            Supports local file path/folder directory, S3 URI and https URL.
        preprocessor_config: The preprocessor config to use for prediction.
        labels: The labels to use for prediction.
        mode: The mode to use for prediction. Can be "first_batch", "average_all", "random_batch", "uniform_batch".
        device: The device to use for prediction. Can be "cpu" or "cuda:0".

    """
    labeled_video_paths = LabeledVideoPaths.from_path(video_or_folder_path)
    labeled_video_paths.path_prefix = ""
    video_sampler = torch.utils.data.SequentialSampler

    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu") if device is None else device

    if mode == "first_batch":
        clip_sampler = pytorchvideo.data.clip_sampling.UniformClipSamplerTruncateFromStart(
            clip_duration=preprocessor_config["clip_duration"],
            truncation_duration=preprocessor_config["clip_duration"],
        )
    elif mode == "average_all":
        clip_sampler = pytorchvideo.data.make_clip_sampler("uniform", preprocessor_config["clip_duration"])
    elif mode == "random_batch":
        clip_sampler = pytorchvideo.data.make_clip_sampler("random", preprocessor_config["clip_duration"])
    elif mode == "uniform_batch":
        clip_sampler = pytorchvideo.data.make_clip_sampler(
            "constant_clips_per_video", preprocessor_config["clip_duration"], 1
        )
    else:
        raise ValueError(
            f"Unknown mode: {mode}. Should be one of 'first_batch', 'average_all', 'random_batch', 'uniform_batch'."
        )

    preprocessor = VideoPreprocessor(**preprocessor_config)
    dataset = LabeledVideoDataset(
        labeled_video_paths,
        clip_sampler,
        video_sampler,
        preprocessor.val_transform,
        decode_audio=False,
        decoder="pyav",
    )

    dataloader = torch.utils.data.DataLoader(dataset, batch_size=1, num_workers=0, drop_last=False)

    results = []
    model = model.to(device)
    model.eval()
    video_index_to_clip_predictions = defaultdict(list)
    video_index_to_video_name = {}
    with torch.inference_mode():
        for batch in dataloader:
            # parse batch
            video_index = batch["video_index"].tolist()[0]
            video_index_to_video_name[video_index] = batch["video_name"][0]
            batch["video"] = batch["video"].to(device)

            # predict
            output = model(batch["video"])

            # apply output activation function
            if model.task == "single_label_classification":
                probabilities = torch.nn.functional.softmax(output, dim=1)
            elif model.task == "multi_label_classification":
                probabilities = torch.nn.functional.sigmoid(output)
            else:
                raise ValueError(f"Unknown task: {model.task}")

            # save predictions
            video_index_to_clip_predictions[video_index].append(probabilities)

    # prepare results
    for video_index, clip_predictions in video_index_to_clip_predictions.items():
        video_name = video_index_to_video_name[video_index]
        probabilities = torch.mean(torch.stack(clip_predictions), dim=1)
        video_predictions = zip(labels, probabilities.tolist()[0])
        video_results = {"filename": video_name, "predictions": dict(video_predictions)}
        results.append(video_results)

    if g_pathmgr.isfile(video_or_folder_path):
        results = results[0]

    return results



import tempfile
from pathlib import Path
from typing import Optional

import torch

from video_transformers.utils.imports import check_requirements


def export(
    model,
    quantize: bool = False,
    opset_version: int = 12,
    export_dir: str = "runs/exports/",
    export_filename: str = "model.onnx",
):
    """
    Exports a model to ONNX format.

    Args:
        model (VideoClassificationModel): The model to export.
        quantize (bool): Whether to quantize the model.
        opset_version (int): The ONNX opset version.
        export_dir (str): The directory to export the model to.
        export_filename (str): The filename to export the model to.
    """
    check_requirements(["torch", "onnx"])

    import onnx

    model.eval()

    Path(export_dir).mkdir(parents=True, exist_ok=True)
    export_path = Path(export_dir) / export_filename
    print(f"Export Path: {export_path}")
    dynamic_axes = {
        "input_data": {0: "batch", 1: "timestamp", 3: "height", 4: "width"},  # write axis names
        "preds": {0: "batch"},
    }
    input_names = ["input_data"]
    output_names = ["preds"]

    if quantize:
        check_requirements(["onnxruntime"])

        from onnxruntime.quantization import quantize_dynamic

        export_filename = Path(export_path).stem + f"_quantized.{Path(export_path).suffix}"

        target_model_path = Path(export_path).parent / export_filename

        with tempfile.NamedTemporaryFile(suffix=".onnx") as temp:
            torch.onnx.export(
                model,
                f=temp.name,
                args=model.example_input_array,
                opset_version=opset_version,
                input_names=input_names,
                output_names=output_names,
                dynamic_axes=dynamic_axes,
                export_params=True,
            )
            quantize_dynamic(temp.name, target_model_path)
    else:
        target_model_path = export_path
        torch.onnx.export(
            model,
            f=target_model_path,
            args=model.example_input_array,
            opset_version=opset_version,
            input_names=input_names,
            output_names=output_names,
            dynamic_axes=dynamic_axes,
            export_params=True,
        )

    onnx_model = onnx.load(target_model_path)
    meta = onnx_model.metadata_props.add()
    meta.key = "labels"
    meta.value = "\n".join(model.labels)
    onnx.save(onnx_model, target_model_path)
    print("Model saved")
